/**
FAT Filesystem Traversal routines

Orion Sky Lawlor, olawlor@acm.org, 2006/04/02 (Public Domain)
*/
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <vector>
#include "fat_layout.h"

/* If set to 1, prints out lots of debugging information */
int verbose=1;


fat_consumer::~fat_consumer() { }

fat_filesystem::fat_filesystem() {fat=NULL;}
fat_filesystem::~fat_filesystem() {delete[] fat;}

/* Read the boot sector and FAT */
void fat_filesystem::setup(void) {
	/* The boot sector is the first thing in the disk */
	read_disk(&boot_sector,sizeof(boot_sector),0);
	
	/* Figure out the sizes of various pieces */
	sector_size=boot_sector.sector_size;
	cluster_size=boot_sector.sec_per_clus*sector_size;
	off_t fat_start=boot_sector.reserved*sector_size;
	off_t fat_size=boot_sector.fat32_length*sector_size;
	fat_count=fat_size/sizeof(fat_t);
	cluster_start=fat_start+fat_size*boot_sector.fats;
	root_cluster=boot_sector.root_cluster;
	cluster_start-=root_cluster*cluster_size; /* silly-- first cluster on disk is "root_cluster", not 0 */
	
	/* Read the FAT */
	fat=new fat_t[fat_count];
	read_disk(&fat[0],fat_size,fat_start);
	
	/* Read the second FAT and check its entries */
	fat_t *f2=new fat_t[fat_count];
	read_disk(&f2[0],fat_size,fat_start+fat_size);
	for (int i=0;i<fat_count;i++)
		if (fat[i]!=f2[i]) 
			abort("First and second FAT entries don't match!\n");
	delete[] f2;
}

/** Something went horribly wrong-- leave. */
void fat_filesystem::abort(const char *why) {
	std::cout<<"Fatal error:"<<why<<"\n";
	std::exit(1);
}

/* Read blocks from this file, and keep passing
   them to this consumer until it returns false (or EOF). */
void fat_filesystem::read_file(cluster_t first_cluster,fat_consumer &c)
{
	if (verbose) std::cout<<"read_file("<<first_cluster<<")\n";
	cluster_t cur=first_cluster;
	while (!fat_islast(cur)) {
		enum {buf_size=512};
		byte buf[buf_size];
		off_t read_start=cluster_start+cluster_size*cur;
		for (int read_size=0;read_size<cluster_size;read_size+=buf_size)
		{ /* Read one buffer of data from disk and consume it */
			read_disk(buf,buf_size,read_start+read_size);
			if (!c.eat_sector(buf,buf_size)) return;
		}
		
		if (verbose) std::cout<<"    cluster "<<cur<<" -> "<<(int)fat[cur]<<"\n";
		/* Advance to the next disk cluster */
		cur=fat[cur];
	}
}

/********************* Directories **********************/

void printDosDir(msdos_dir_entry &e) {
	char dosName[12];
	std::strncpy(dosName,(char *)e.name,11);
	dosName[11]=0;
	std::cout<<"  DOS: name='"<<dosName<<"'  ";
	
	std::cout<<"  attr="<<(int)e.attr<<", "
	    "cluster="<<e.start_cluster()<<","
	    "size="<<(int)e.size<<"\n";
}
/* Convert UNICODE little-endian 16-bit characters to std::string */
std::string winName(const byte *data,int nChar) {
	std::string ret;
	for (int i=0;i<nChar;i++) ret+=(char)data[2*i+0];
	return ret;
}
void printWinDir(win_dir_entry &e) {
	std::cout<<"  WIN: name='"<<
		winName(e.name0_4,5)+winName(e.name5_10,6)+winName(e.name11_12,2)
		<<"'  id="<<(int)e.id<<"\n";
}

class dir_printer : public fat_consumer { 
public:
	bool eat_sector(void *data,int len) {
		msdos_dir_entry *e=(msdos_dir_entry *)data;
		int n=len/sizeof(msdos_dir_entry);
		std::cout<<" Directory sector:\n";
		for (int i=0;i<n;i++) {
			if (e[i].attr&ATTR_VOLUME) /* volume label bit set-- windows long filename */
				printWinDir(*(win_dir_entry *)&e[i]);
			else /* normal DOS directory entry */
				printDosDir(e[i]);
		}
		return true; /* give me more sectors! */
	}
};



/***************** Caller code ******************/
class fstream_fat : public fat_filesystem {
	std::ifstream &fs;
public:
	fstream_fat(std::ifstream &fs_) :fs(fs_) {}
	virtual void setup(void) {
		fat_filesystem::setup();
		if (verbose) {
			std::cout<<"FAT filesystem layout:\n";
			std::cout<<"  FAT consists of "<<fat_count<<" entries\n";
			std::cout<<"  Each entry is for one cluster of "<<cluster_size<<" bytes\n";
			std::cout<<"  Clusters start on disk at byte "<<cluster_start<<"\n";
			std::cout<<"  First few FAT entries: \n";
			for (int i=0;i<20;i++) std::printf("      fat[%X]=%X \n",i,(int)fat[i]);
			std::cout<<"\n\n";
		}
	}
	virtual void read_disk(void *p,unsigned int n,off_t o) {
		if (verbose) std::cout<<"       read_disk("<<p<<","<<n<<","<<o<<");\n";
		fs.seekg(o);
		fs.read((char *)p,n);
		if (!fs) abort("unable to read disk");
	}
};



int main(int argc,char *argv[]) {
	if (argc<=1) { /* Not enough command-line arguments */
		std::cout<<"Usage: fat <FAT32 filesystem image>\n"; 
		std::exit(1);
	}
	std::ifstream diskImage(argv[1]);
	fstream_fat fs(diskImage);
	fs.setup();
	
	/* Print out contents of root directory */
	std::cout<<"Contents of root directory:\n";
	dir_printer dp;
	fs.read_file(fs.root_cluster,dp); /* FIXME: does root directory always start at cluster 0? */
	
	return 0;
}
